import bw_processing as bwp
import pandas as pd
import numpy as np
from .metadata_manager import *


class UncertaintyHandler:
    def __init__(self):
        self.metadata = metadata_manager._get_metadata()

    def _calc_specific_uncertainty(self, data, uncertainty):
        loc = np.log(data)
        scale = np.log(uncertainty)

        return loc, scale

    def _generate_uncertainty_tuple(self, data, type, gsd, uncertainty_negative):
        """
        Generate the uncertainty tuple for one value.

        Parameters:
            * data: The input or output value of the one exchange in the system.
            * type: The type of uncertainty, such as 2.
            * gsd: Geometric Standard Deviation, used to calculate sigma of lognormal distribution.
            * uncertainty_negative: uncertainty negative
        """
        data = abs(data)
        if type in [0, 1]:
            uncertainty_tuple = (type, data, np.NaN, np.NaN, np.NaN, np.NaN, uncertainty_negative)
        elif type == 2:  # no need to consider data == 0, because sparse matrix only save non-zero values.
            if gsd == 0:
                uncertainty_tuple = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, uncertainty_negative)
            else:
                uncertainty_tuple = (type, np.log(data), np.log(gsd), np.NaN, np.NaN, np.NaN, uncertainty_negative)
        elif type == 3: # normal
            uncertainty_tuple = (type, data, np.NaN, np.NaN, np.NaN, np.NaN, uncertainty_negative)
        elif type == 4: # uniform
            uncertainty_tuple = (type, np.NaN, np.NaN, np.NaN, (data - data * gsd), (data + data * gsd), uncertainty_negative)

        return uncertainty_tuple
    
    def _get_uncertainty_value(self, strategy, act_index, row):
        """
        Get uncertainty values by strategy from metadata.

        Parameters:
            * strategy: The strategy of adding uncertainty, "itemwise" or "columnwise".
            * act_index: The index of the activity.
            * row: The row number of the corresponding activity.
        """
        if self.metadata is None:
            print("Please write your uncertainty information into metadata first.")
        
        if len(self.metadata) <= row:  # If biosphere, indices need to be subtracted from technosphere indices.
            row = row - len(self.metadata) + 1

        if strategy == "itemwise":
            specific = self.metadata[act_index].get("Exchange uncertainty amount", 0)
            if isinstance(specific, list):
                uncertainty_value = specific[row]
            else:  # because some activity don't have uncertainty at all, in this case is background system
                uncertainty_value = specific
        elif strategy == "columnwise":
            uncertainty_value = self.metadata[act_index].get("Exchange uncertainty amount", 0)
            uncertainty_value = 0 if np.isnan(uncertainty_value) else uncertainty_value

        return uncertainty_value
    
    def _get_uncertainty_negative(self, strategy, act_index, row):
        """
        Get uncertainty negative (TRUE or FALSE) by strategy from metadata.

        Parameters:
            * strategy: The strategy of adding uncertainty, "itemwise" or "columnwise" (If strategy is "Uniform", don't need to call this function).
            * act_index: The index of the activity.
            * row: The row number of the corresponding activity.
        """
        if self.metadata is None:
            print("Please write your uncertainty information into metadata first.")
        
        if len(self.metadata) <= row:  # If biosphere, indices need to be subtracted from technosphere indices.
            row = row - len(self.metadata) + 1

        if strategy == "itemwise":
            specific = self.metadata[act_index].get("Exchange negative", False)
            if isinstance(specific, list):
                uncertainty_negative = specific[row]
            else:  # because some activity don't have uncertainty at all, in this case is background system
                uncertainty_negative = specific
        elif strategy == "columnwise":
            uncertainty_negative = self.metadata[act_index].get("Exchange negative", False)

        return uncertainty_negative

    def add_nonuniform_uncertainty(self, bw_data, bw_indices, bg_strategy, fg_num=None, fg_strategy=None):
        """
        Prepare uncertainty array for datapackage. By default, foreground system is not considered, but you can set youself.

        Parameters:
            * bw_data: One of the technosphere/bioshphere/characterization factor matrix in datapackage needed format.
            * bw_indices: Matrix indices in datapackage needed format.
            * bw_flip: Flip array of technosphere in datapackage needed format.
            * bg_strategy: The uncertainty stragegy for one matrix in background system, two options available: "itemwise" and "columnwise"
            * fg_num: The number of columns in the foreground system.
            * fg_strategy: The uncertainty stragegy fsor one matrix in foreground system, two options available: "itemwise" and "columnwise"
        """
        if self.metadata is None:
            print("Please write your uncertainty information into metadata first.")

        uncertainty_array = []
        for i, data in enumerate(bw_data):
            row, col = bw_indices[i]
            act_index = col
            uncertainty_type = self.metadata[act_index].get("Activity uncertainty type", 0)
            if fg_num is not None and fg_strategy is not None:
                if act_index < fg_num: # foreground situation
                    strategy = fg_strategy
                else:
                    strategy = bg_strategy
            else:
                strategy = bg_strategy
            
            uncertainty_array.append(self._generate_uncertainty_tuple(data, uncertainty_type, self._get_uncertainty_value(strategy, act_index, row), self._get_uncertainty_negative(strategy, act_index, row)))

        return np.array(uncertainty_array, dtype=bwp.UNCERTAINTY_DTYPE)
    
    def add_uniform_uncertainty(self, type, gsd, uncertainty_negative, bw_data, bw_flip=None):
        """
        Generate the uncertainty tuple for one value.

        Parameters:
            * type: The type of uncertainty, such as 0, 1, 2, 3, 4.
            * gsd: Geometric Standard Deviation, used to calculate sigma of lognormal distribution.
            * bw_data: All values for foreground system or background system.
            * bw_flip: The flip array of technosphere.
        """
        uncertainty_array = []

        if bw_flip is not None:
            for i in range(len(bw_data)):
                if bw_flip[i] == True:
                    uncertainty_array.append(self._generate_uncertainty_tuple(bw_data[i], type, gsd, uncertainty_negative))
                else:
                    uncertainty_array.append((0, bw_data[i], np.NaN, np.NaN, np.NaN, np.NaN, False))
        else:
            for i in range(len(bw_data)):
                uncertainty_array.append(self._generate_uncertainty_tuple(bw_data[i], type, gsd, uncertainty_negative))

        return np.array(uncertainty_array, dtype=bwp.UNCERTAINTY_DTYPE)


import pandas as pd
from .metadata_manager import *


class UncertaintyImporter:
    """
    The metadata updated in this class will be stored in metadata_manager
    """
    def __init__(self, file_path, delimiter):
        self.metadata = metadata_manager._get_metadata()
        self.file_path = file_path
        self.delimiter = delimiter
        self.df = None

    def _load_df(self):
        if self.df is None:
            self.df = pd.read_csv(self.file_path, delimiter=self.delimiter)

    def update_metadata_uncertainty(self, activities, strategy):  # TODO: What if they have the same name?
        """
        Parameters:
            * strategy: "itemwise" or "columnwise"
        """
        self._update_metadata_activities(activities)

        self._load_df()
        column_names = self.df["Activity name"].unique()  # the column names of foreground system
        for col_name in column_names:
            self._update_metadata_column_uncertainty(col_name, self.df, activities, strategy)

    def _update_metadata_activities(self, activities):
        """
        Import all foreground activity names, if foreground system has 2 columns, it will have 2 keys.
        """
        if not any("Activity name" in value for value in self.metadata.values()):
            for i in range(len(activities)):
                metadata_manager._update_metadata(i, {"Activity name": activities[i]})

    def _update_metadata_column_uncertainty(self, act_name, df, activities, strategy):
        """
        Update metadata by "Activity name"
        """
        selected_df = df.loc[df["Activity name"] == act_name, ["Exchange name", "Exchange uncertainty type", "GSD", "Exchange negative"]]
        selected_dict = selected_df.set_index("Exchange name")[["Exchange uncertainty type", "GSD", "Exchange negative"]].to_dict(orient="index")

        if strategy == "itemwise":
            selected_df_2 = selected_df.set_index("Exchange name")[["Exchange uncertainty type", "GSD", "Exchange negative"]]
            selected_df_2 = selected_df_2.reindex(activities, fill_value=0)
            gsd_list = selected_df_2["GSD"].astype(float).fillna(0).tolist()
            negative_list = selected_df_2["Exchange negative"].replace(0, False).astype(bool).fillna(False).tolist()
            for key, value in self.metadata.items():
                activity_name = value["Activity name"]
                if activity_name in selected_dict:
                    index = activities.index(act_name)
                    uncertainty_type = selected_df[selected_df["Exchange name"] == act_name]["Exchange uncertainty type"].to_list()[0]
                    self.metadata[index]["Activity uncertainty type"] = uncertainty_type
                    self.metadata[index]["Exchange uncertainty amount"] = gsd_list
                    self.metadata[index]["Exchange negative"] = negative_list
        elif strategy == "columnwise":
            for key, value in self.metadata.items():
                activity_name = value["Activity name"]
                if activity_name in selected_dict:
                    index = activities.index(act_name)
                    uncertainty_type = selected_df[selected_df["Exchange name"] == act_name]["Exchange uncertainty type"].to_list()[0]
                    self.metadata[index]["Activity uncertainty type"] = uncertainty_type
                    self.metadata[index]["Exchange uncertainty amount"] = selected_df[selected_df["Exchange name"] == act_name]["GSD"].to_list()[0]
                    self.metadata[index]["Exchange negative"] = selected_df[selected_df["Exchange name"] == act_name]["Exchange negative"].to_list()[0]
        else:
            print(F"Strategy {strategy} is not supported, you should either choose 'columnwise' or 'itemwise'")


import bw_processing as bwp
import pandas as pd
import numpy as np


def detect_foreground(acts, bg_activities):
    """
    Detect foreground activties from user's input.
    """
    fg_activities = list(set(acts) - set(bg_activities))

    return fg_activities

def get_country_sector(activity):
    """
    Design for EXIOBASE: Separate the country and sector.
    
    Parameters:
        * activity: The activity name that needs to be processed.
    """
    country, sector = activity.split("-", 1)

    return country, sector

def get_fg_dataframe(user_dataframe: pd.DataFrame, fg_activities: list):
    """
    Extract the foreground rows from user's input file.
    
    Parameters:
        * user_dataframe: The dataframe of user's input file.
        * fg_activities: All foreground activities.
    """
    fg_dataframe = user_dataframe[user_dataframe["Activity name"].isin(fg_activities)]

    return fg_dataframe

def get_fg_activities(fg_file_path: str, delimiter: str, bg_activities: list) -> list:
    """
    Get all activities of foreground system.

    Parameters:
        * fg_file_path: The path to the file that needs to be processed.
        * delimiter: The separator of the file.
    """
    df = pd.read_csv(fg_file_path, decimal=delimiter)
    acts = df["Activity name"].unique().tolist()
    fg_activities = detect_foreground(acts, bg_activities)

    return fg_activities

def get_bg_activities(a_file_path: str, delimiter: str) -> list:
    """
    Get all activities of EXIOBASE by combing <country_name> and <sector_name>.

    Parameters:
        * a_file_path: The path to the file that needs to be processed.
        * delimiter: The separator of the file.
    """
    df = pd.read_csv(a_file_path, delimiter=delimiter, header=None)
    countries = df.iloc[3:, 0].unique().tolist()
    sectors = df.iloc[3:, 1].unique().tolist()
    activities = [ x + '-' + y for x in countries for y in sectors]

    return activities

def file_preprocessing(file_name, delimiter: str, column_name: str, expacted_order: list):
    """
    Preprocess a file and return a DataFrame with the desired order.

    Parameters:
        * file_name: The path to the file that needs to be processed.
        * delimiter: The delimiter used in the file, for example: ','.
        * column_name: The column name of unexpected order.
        * expected_order: A list specifying the desired order of the rows in the DataFrame.
    """
    df = pd.read_csv(file_name, delimiter=delimiter)
    df_sorted = df.set_index(column_name).reindex(expacted_order).reset_index()

    return df_sorted

def map_pedigree_uncertainty(country_file, sector_file, region_sector_file):
    """
    Design for case study: Build dictionaries to mapping specific uncertainty.
    
    Parameters:
        * country_file: The file that group country to region.
        * sector_file: The file that group sector to aggregated sector.
        * region_sector_file: The mapping from aggregated region and sector to GSD. 
    """
    country_data = pd.read_csv(country_file, delimiter=";").fillna("").to_numpy()
    sector_data = pd.read_csv(sector_file, delimiter=";").fillna("").to_numpy()
    region_sector_data = pd.read_csv(region_sector_file, delimiter=";").fillna("").to_numpy()

    country_region = {row[0]: row[1] for row in country_data}
    sector_seccat = {row[0]: row[1] for row in sector_data}

    return country_region, sector_seccat, region_sector_data

def find_pedigree_uncertainty(activity, country_region, sector_seccat, region_sector_dfs):
    """
    Design for case study: Search for pedigree uncertainty for specific activity or biosphere flow.
    """
    country, sector = get_country_sector(activity)
    region_category =  country_region.get(country, None)
    sector_category = sector_seccat.get(sector, None)

    if region_category !=  None and sector_category != None:
        gsd = float(region_sector_dfs[(region_sector_dfs.iloc[:, 0] == region_category) & (region_sector_dfs.iloc[:, 1] == sector_category)]["GSD"].iloc[0])
    else:
        print("No GSD found.")
        gsd = None

    return gsd

def add_uncertainty(self, bw_data, bw_indices, bw_flip):
    """
    Design for case study: Add pedigree and specific uncertainty
    """
    bw_uncertainties = []
    if bw_flip is not None: # technosphere
        k = 0
        for data, indices, flip in zip(bw_data, bw_indices, bw_flip):
            uncertainty = list(self.metadata[indices[1]].items())[0][1][1]
            if uncertainty is not None:
                if indices[1] == 0:
                    uncertainty = uncertainty[k]
                    k += 1
                    if uncertainty == 0 or data == 0:
                        parameters_a = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, False)
                    else:
                        loc = np.log(abs(data))
                        scale = np.log(uncertainty)
                        if not flip:
                            parameters_a = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, False)
                        else:
                            parameters_a = (2, loc, scale, np.NaN, np.NaN, np.NaN, False)
                else:
                    if data == 0:
                        parameters_a = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, False)
                    else:
                        loc = np.log(abs(data))
                        scale = np.log(uncertainty)
                        if not flip:
                            parameters_a = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, False)
                        else:
                            parameters_a = (2, loc, scale, np.NaN, np.NaN, np.NaN, False)
            else:
                parameters_a = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, False)
            bw_uncertainties.append(parameters_a)
    else:
        k = 0
        for data, indices in zip(bw_data, bw_indices):
            uncertainty = list(self.metadata[indices[1]].items())[0][1][1]
            if uncertainty is not None:
                if indices[1] == 0:
                    uncertainty = uncertainty[k]
                    k += 1
                    if uncertainty == 0 or data == 0:
                        parameters_a = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, False)
                    else:
                        loc = np.log(abs(data))
                        scale = np.log(uncertainty)
                        parameters_a = (2, loc, scale, np.NaN, np.NaN, np.NaN, False)
                else:
                    if data == 0:
                        parameters_a = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, False)
                    else:
                        loc = np.log(abs(data))
                        scale = np.log(uncertainty)
                        parameters_a = (2, loc, scale, np.NaN, np.NaN, np.NaN, False)
            else:
                parameters_a = (0, data, np.NaN, np.NaN, np.NaN, np.NaN, False)
            bw_uncertainties.append(parameters_a)

    return np.array(bw_uncertainties, dtype=bwp.UNCERTAINTY_DTYPE)


import bw2calc as bc
import pandas as pd
import numpy as np
import os


class LCAWrapper:
    def perform_static(self, demand, datapackage, directory, k, t, myact):
        """
        Perform static simulation and save the lca score.

        Parameters:
            * demand: The dictionary of the functional unit, for example: {<index>: <amount>}.
            * datapackage: The datapackage used for simulation.
            * directory: The directory to save output file.
            * k: The case identifier.
            * t: The type of the simulation, such as "static", "uniform_0.2".
            * myact: 
        """
        lca = bc.LCA(
            demand=demand,
            data_objs=[datapackage],
        )
        lca.lci()
        lca.lcia()

        print(f"Brightway calculated lca score: {lca.score, myact}")
        os.makedirs(directory, exist_ok=True)
        filename = os.path.join(directory, f"CASE_{k}_{t}_MC_simulations_{myact}.csv")

        with open(filename, "w") as file:
            file.write("kg CO2eq\n") # TODO: Write the header -> Not use this function, so it's ok for now.
            file.write(f"{lca.score}")
            print(f"Static LCA result saved to {filename}.")

    def perform_stochastic(self, index, datapackage, directory, k, t, myact, batch_size=50, num_batches=10):
        """
        Perform Monte Carlo simulation and save the lca score.
        """
        lca = bc.LCA(
            demand={index: 1},
            data_objs=[datapackage],
            use_distributions=True,
        )
        lca.lci()
        lca.lcia()

        print(f"Brightway calculated lca score(with uncertainty): {lca.score, myact}")
        os.makedirs(directory, exist_ok=True)
        filename = os.path.join(directory, f"CASE_{k}_{t}_MC_simulations_{myact}.csv")

        with open(filename, "w") as file:
            file.write("kg CO2eq\n")
            for p in range(num_batches):
                batch_results = [lca.score for _ in zip(range(batch_size), lca)]
                df_batch = pd.DataFrame(batch_results, columns=["kg CO2eq"])
                df_batch.to_csv(file, header=False, index=False)
                print(f"Batch {p} saved to {filename}.")

        print(f"Results saved to {filename}.")

    def manual_lca(self, A, B, C, index): 
        """
        Perform Monte Carlo simulation without brightway.
        """
        f = np.zeros((len(A), 1))
        f[index] = 1
        lca_score = np.sum(C.dot(B.dot((np.linalg.inv(A)).dot(f))))
        
        return float(lca_score)


import pandas as pd
import numpy as np


class ForegroundImporter:
    def extend_matrix(self, extend_data: pd.DataFrame, emissions: list, fg_activities: list, bg_activities: list):
        """
        Concatenate foreground data to background data.
        
        Parameters:
            * extend_data: The user's input data include technosphere and biosphere data in dataframe format.
            * emissions: The list of emissions.
            * fg_activities: The list of foreground activities.
            * bg_activities: The list of background activities.
        """
        fgbg = np.zeros([len(fg_activities), len(bg_activities)])
        fgfg = np.zeros([len(fg_activities), len(fg_activities)])
        bgfg = np.zeros([len(bg_activities), len(fg_activities)])
        bifg = np.zeros([len(emissions), len(fg_activities)])

        for index, row in extend_data.iterrows():
            activity_name = row["Activity name"]
            column_num = fg_activities.index(activity_name)

            if row["Exchange type"] == "production":  # fgfg
                row_num = fg_activities.index(row["Exchange name"])
                fgfg[row_num][column_num] = row["Exchange amount"]
            elif row["Exchange type"] == "technosphere": # bgfg
                try:
                    row_num = bg_activities.index(row["Exchange name"])
                except ValueError:
                    row_num = fg_activities.index(row["Exchange name"])
                bgfg[row_num][column_num] = row["Exchange amount"]
            elif row["Exchange type"] == "biosphere": # bifg
                row_num = emissions.index(row["Exchange name"])
                bifg[row_num][column_num] = row["Exchange amount"]

        fgfg = np.nan_to_num(fgfg, nan=0)
        bgfg = np.nan_to_num(bgfg, nan=0)
        bifg = np.nan_to_num(bifg, nan=0)

        return fgbg, fgfg, bgfg, bifg

    def concatenate_matrix(self, tech_matrix: pd.DataFrame, bio_matrix: pd.DataFrame, fgbg: np.ndarray, fgfg: np.ndarray, bgfg: np.ndarray, bifg: np.ndarray):
        tech_matrix = np.concatenate((np.concatenate((fgfg, bgfg), axis=0), np.concatenate((fgbg, tech_matrix), axis=0)), axis=1)
        bio_matrix = np.concatenate((bifg, bio_matrix), axis=1)

        return tech_matrix, bio_matrix



import bw_processing as bwp
from scipy import sparse
import pandas as pd
import numpy as np
from typing import List, Tuple, Any


class DatapackageBuilder:
    def prepare_dp_matrix(self, tech_matrix, bio_matrix, cf_matrix):
        """
        Transform matrices data to bw matrices data, ready for the datapackages.
        """
        tech_sparse = sparse.coo_array(tech_matrix)
        tech_coors = np.column_stack(tech_sparse.nonzero())
        
        max_coor = tech_coors[np.argmax(np.sum(tech_coors, axis=1))]
        tech_data = tech_sparse.data
        tech_indices = np.array([tuple(coor) for coor in tech_coors], dtype=bwp.INDICES_DTYPE)
        tech_flip = np.array([False if i[0] == i[1] else True for i in tech_indices])

        bio_sparse = sparse.coo_array(bio_matrix)
        bio_coors = np.column_stack(bio_sparse.nonzero())
        bio_data = bio_sparse.data
        bio_indices = np.array([tuple([coord[0] + max_coor[0] + 1, coord[1]]) for coord in bio_coors], dtype=bwp.INDICES_DTYPE)
        
        cf_sparse = sparse.coo_array(cf_matrix)
        cf_coors = np.column_stack(cf_sparse.nonzero())
        cf_data =  cf_sparse.data
        cf_indices = np.array([tuple([coord[0] + max_coor[0] + 1, coord[1] + max_coor[1] + 1]) for coord in cf_coors], dtype=bwp.INDICES_DTYPE)

        return [
            (tech_data, tech_indices, tech_flip),
            (bio_data, bio_indices),
            (cf_data, cf_indices)
        ]

    def prepare_datapackage(self, datapackage_data: List[Tuple[Any, ...]], uncertainty: list = None):
        """
        Prepare datapackage for brightway LCA calculation.

        Parameters:
            * datapackage_data: A list of tuple includes all information to create a datapackage.
            * uncertainty: The uncertainty for all matrices.
        """
        tech_data, tech_indices, tech_flip = datapackage_data[0]
        bio_data, bio_indices = datapackage_data[1]
        cf_data, cf_indices = datapackage_data[2]
        if uncertainty is None:
            tech_uncertainty, bio_uncertainty, cf_uncerainty = None, None, None
        else:
            tech_uncertainty, bio_uncertainty, cf_uncerainty = uncertainty[0], uncertainty[1], uncertainty[2]

        dp = bwp.create_datapackage()
        dp.add_persistent_vector(
            matrix='technosphere_matrix',
            indices_array=tech_indices,
            data_array=tech_data,
            flip_array=tech_flip,
            distributions_array=tech_uncertainty,
        )
        dp.add_persistent_vector(
            matrix='biosphere_matrix',
            indices_array=bio_indices,
            data_array=bio_data,
            distributions_array=bio_uncertainty,
        )
        dp.add_persistent_vector(
            matrix='characterization_matrix',
            indices_array=cf_indices,
            data_array=cf_data,
            distributions_array=cf_uncerainty,
        )

        return dp


class MetadataManager:
    """
    This class is used to manage metadata of this library, the metadata is designed for handling ununiformly uncertainty analysis.

    The metadata with uncertainty information follows the following structure:

        self.metadata: save activity information, including: 
                activity index(the index in the datapackage), 
                uncertainty type, 
                activity name, 
                itemwise uncertainty value,
                and columnwise uncertainty value. (Here the uncertainy value means the mu value.)

        metadata = {
            index1: {
                "type": 1,
                "act": act1,
                "specific": [specific1, specific2, specific3, ...],
                "gsd": 0,
            },
            index2: {
                "type": 2,
                "act": act2,
                "specific": [],
                "gsd": gsd2,
            },
            ...
        }
        """
    _instance = None
    
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        if not hasattr(self, "metadata"):
            self.metadata = {}

    def _update_metadata(self, key, value):
        self.metadata[key] = value

    def _get_metadata(self):
        return self.metadata
    
metadata_manager = MetadataManager()


import pandas as pd
import numpy as np
import bw2data as bd
from .utils import *


class BackgroundImporter:
    def build_tech_matrix(self, raw_tech: np.ndarray):
        """
        Get technosphere matrix data:
            Calculate (I-A), then convert it into a totally positive matrix.

        Parameters:
            * raw_tech: Raw data in pandas dataframe format.

        Returns: 
            * np.ndarray: Return the numpy matrix format technosphere.
        """
        identity_matrix = np.identity(len(raw_tech))
        tech_matrix = - (identity_matrix - raw_tech)
        np.fill_diagonal(tech_matrix, -tech_matrix.diagonal())

        return tech_matrix

    def build_bio_matrix(self, bio_df, emissions) -> np.ndarray:
        """
        Get biosphere matrix data:
            Extract the corresponding value by emission name.
        
        Parameters:
            * bio_df: The whole EXIOBASE emissions in dataframe format.
            * emissions: The list of emissions used for LCA calculation.

        Returns: 
            * np.ndarray: Return the numpy matrix format biosphere.
        """
        bio_matrix = bio_df.loc[emissions].to_numpy()

        return bio_matrix

    def _get_from_cfs(self, emission_df):
        """
        Get the characterization factors (type: list) from characterization factor file in dataframe format.

        Parameters:
            * emission_df: The dataframe format of characterization factor file, provide characterization factor matrix raw data.
            
        Returns:
            * list: Return the CFs as a list.
        """
        column_name = "CFs"
        if column_name not in emission_df.columns:
            print("The file doesn't have a 'CFs' column. Please add your characterization factor values to the 'CFs' column.")
        else:
            cf_df = emission_df[column_name]
            if not cf_df.isnull().any():
                print("All characterization factors have been found.")
                return cf_df.to_list()
            else:
                emission_names = emission_df["exiobase name"].to_list()
                emission_values = emission_df[column_name].to_list()
                cf_missing = emission_df[emission_df[column_name].isnull()]["exiobase name"].to_list()
                
                print(f"{cf_missing} emission(s) don't have characterization factor values. Please complete your 'CFs' column.")
                # print the founded CFs
                # print(f"Emissions: \n{emission_names}")
                # print(f"Characteriation factor values: \n{emission_values}")

        return None

    def _get_from_code(self, emission_df, method, ecoinvent_name):
        """
        Get the characterization factor values (type: list) through brightway by code.

        Parameters:
            * emission_df: The dataframe format of characterization factor file, provide characterization factor matrix raw data.
            * method: The selected method used for LCA calculation.
            * ecoinvent_name: The name of the ecoinvent database on user's device.

        Returns:
            * list | None: A list of characterization factor (cf) values if found, otherwise None.
        """

        emission_codes = emission_df["brightway code"].copy()
        emission_names = emission_df["exiobase name"].copy()
        code_name = pd.concat([emission_codes, emission_names], axis=1, ignore_index=True)
        code_name.columns = ["code", "name"]

        bw_method = bd.Method(method)
        method_data = bw_method.load() # method_data is a list of tuple

        if isinstance(method_data[0][0], list): # for ecoinvent 3.9, the first element of the tuple is a two element list
            method_df = pd.DataFrame(method_data, columns=["database_code", "cf_value"])
            method_df[["database", "code"]] = method_df["database_code"].to_list()
            cf_selected = method_df[method_df["code"].isin(emission_codes)][["code", "cf_value"]].copy()
            missing_codes = list(set(emission_codes.unique()) - set(cf_selected["code"]))

            code_name_cf_value = code_name.merge(cf_selected, on="code", how="left", suffixes=("_old", "_new"))
            cf_values = code_name_cf_value["cf_value"]
            if missing_codes:
                emission_names = []
                for missing_code in missing_codes:
                    emission_names.append(emission_df[emission_df["brightway code"] == missing_code]["exiobase name"].to_list()[0])
                print(f"Characterization factor data incomplete, missing: {list(zip(emission_names, missing_codes))}")
            else:
                print("All characterization factors have been found.")
                return cf_values.to_list()
        elif isinstance(method_data[0][0], int): # for ecoinvent 3.11, the first element of the tuple is an int value
            method_df = pd.DataFrame(method_data, columns=["id", "cf_value"])
            emissions = {}
            for code in emission_codes:
                emission = bd.Database(ecoinvent_name).get(code)
                emissions[emission.id] = code  # create the mapping from id to code
            ids = list(emissions.keys())
            cf_selected = method_df[method_df["id"].isin(ids)][["id", "cf_value"]]
            missing_ids = set(ids) - set(cf_selected["id"].to_list())
            missing_codes = [emissions[key] for key in missing_ids if key in emissions]
            emissions = pd.DataFrame(list(emissions.items()), columns=["id", "code"])

            name_code_id = code_name.merge(emissions, how="left", on="code", suffixes=("_old", "_new"))
            df_merged = name_code_id.merge(cf_selected, how="left", on="id", suffixes=("_old", "_new"))
            matrix_values = df_merged["cf_value"]
            if missing_codes:
                emission_names = []
                for missing_code in missing_codes:
                    emission_names.append(emission_df[emission_df["brightway code"] == missing_code]["exiobase name"].to_list()[0])
                print(f"Characterization factor data incomplete, missing: {list(zip(emission_names, missing_codes))}")
            else:
                print("All characterization factors have been found.")
                return matrix_values.to_list()
        
        return None

    def build_cf_matrix(self, emission_file: str, emission_list: list, biodb_name: str = None, method: tuple = None, source="cf") -> np.ndarray:
        """
        Get characterization factor matrix data.

        Parameters: 
            * emission_file: The path to the file that needs to be processed. The file includes emission name and emission code column.
            * emission_list: the list of emissions in foreground system
            * biodb_name: The name of the biosphere database on user's device.
            * method: The LCIA method used for LCA calculation.
            * source: define the data source, two options: "cf" or "code". Set to "cf", the function extract CFs from "cf value" column of the file, set to "code", the function extract CFs from "brightway code" column of the file.

        Returns: 
            * np.ndarray | None: Return the numpy matrix format of characterization factor (cf) values if found, otherwise None.
        """
        emission_df = pd.read_csv(emission_file, delimiter=",")
        emission_df = file_preprocessing(emission_file, ",", "exiobase name", emission_list)  # sorting the column order align with the desired order.
        
        if source == "cf":
            cf_values = self._get_from_cfs(emission_df)
            if cf_values:
                cf_matrix = np.diagflat(cf_values)
                return cf_matrix
            else:
                print("Failed to build matrix, there are CFs not found.")
        elif source == "code":
            cf_values = self._get_from_code(emission_df, method, biodb_name)
            if cf_values:
                cf_matrix = np.diagflat(cf_values)
                return cf_matrix
            else:
                print("Failed to build matrix, there are CFs not found.")
        else:
            print('Please set the source to either "cf" or "code".')
            
        return None
    
    # TODO: if it's functional unit, then try to find it and set it to 1 -> Not use this function, it's ok for now.
    def _find_functional_unit(self, emission_code, missing_codes, cf_dict, codes):
        cf_matrix = []
        if missing_codes: # try to find co2
            miss_dict = emission_code[["ecoinvent name", "brightway code"]].set_index("brightway code")["ecoinvent name"].to_dict()
            fixed_codes = []
            for code in missing_codes:
                name = miss_dict.get(code)
                if "Carbon dioxide" in name:
                    cf_dict[code] = 1.0
                    fixed_codes.append(code)
            if missing_codes == fixed_codes:
                for code in codes:
                    cf_matrix.append(cf_dict.get(code))

        cf_matrix = np.diagflat(cf_matrix)

        if len(cf_matrix) != len(emission_code):
            raise ValueError(f"Characterization factor data incomplete, missing: {missing_codes}")
            
        return cf_matrix